# Copyright 2018 Descartes Labs.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Displays ndarrays as images, but is easier to use and more flexible than matplotlib's ``imshow``.
"""

from __future__ import division
import six

import descarteslabs.client.addons


def display(*imgs, **kwargs):
    """
    Display 2D and 3D ndarrays as images with matplotlib.

    The ndarrays must either be 2D, or 3D with 1 or 3 bands.
    If they are 3D masked arrays, the mask will be used as an alpha channel.

    Unlike matplotlib's ``imshow``, arrays can be any dtype;
    internally, each is normalized to the range [0..1].

    Parameters
    ----------
    *imgs: 1 or more ndarrays
        When multiple images are given, each is displayed on its own row.
    bands_axis: int, default 0
        Axis which contains bands in each array.
    title: str, or sequence of str; optional
        Title for each image. If a sequence, must be the same length as ``imgs``.
    size: int, default 10
        Length, in inches, to display the longer side of each image.
    robust: bool, default True
        Use the 2nd and 98th percentiles to compute color limits.
        Otherwise, the minimum and maximum values in each array are used.
    interpolation: str, default "bilinear"
        Interpolation method for matplotlib to use when scaling images for display.

        Bilinear is the default, since it produces smoother results when scaling
        down continuously-valued data (i.e. images). For displaying discrete data,
        however, choose 'nearest' to prevent values not existing in the input
        from appearing in the output.

        Acceptable values are 'none', 'nearest', 'bilinear', 'bicubic', 'spline16',
        'spline36', 'hanning', 'hamming', 'hermite', 'kaiser', 'quadric', 'catrom',
        'gaussian', 'bessel', 'mitchell', 'sinc', 'lanczos'

    Example
    =======

    .. ipython:: python

        import descarteslabs as dl
        import numpy as np
        a = np.arange(20*15).reshape((20, 15))
        b = np.tan(a)
        @savefig display.png
        dl.scenes.display(a, b, size=4)

    Raises
    ------
    ImportError
        If matplotlib is not installed.
    """
    if len(imgs) == 0:
        return

    bands_axis = kwargs.pop("bands_axis", 0)
    titles = kwargs.pop("title", None)
    size = kwargs.pop("size", 10)
    robust = kwargs.pop("robust", True)
    interpolation = kwargs.pop("interpolation", "bilinear")
    if len(kwargs) > 0:
        raise TypeError("Unexpected keyword arguments for display: {}".format(', '.join(six.iterkeys(kwargs))))

    np = descarteslabs.client.addons.numpy
    matplotlib = descarteslabs.client.addons.import_matplotlib_pyplot()
    plt = matplotlib.pyplot

    if len(imgs) == 1:
        if isinstance(imgs[0], (list, tuple)):
            raise TypeError("To display a sequence of images, unpack it: `display(*images_list)`")
        elif isinstance(imgs[0], np.ndarray) and len(imgs[0].shape) == 4:
            raise TypeError("To display a 4D ndarray (image stack), unpack it: `display(*stack)`")

    # TODO: facet grid
    # TODO: leaves huge gaps between images that aren't very square
    # would need to calculate figsize better based on shapes of each img
    # or use seaborn?
    figsize = (size, size * len(imgs))
    fig, axs = plt.subplots(len(imgs), 1, figsize=figsize, squeeze=False)

    if isinstance(titles, (list, tuple, np.ndarray)):
        if len(titles) != len(imgs):
            raise ValueError("Different number of titles given than images")
    else:
        titles = [titles] * len(imgs)

    for ax, img, title in zip(axs, imgs, titles):
        ax = ax[0]

        if not isinstance(img, np.ndarray):
            raise TypeError("Expected ndarray, instead got {}".format(type(img)))

        if len(img.shape) not in (2, 3):
            raise NotImplementedError(
                "Can only display 2D or 3D arrays, not shape {}".format(img.shape)
            )

        if len(img.shape) == 2:
            # expand 2d image to 3d with 1 band
            slicer = [slice(None)] * 3
            slicer[bands_axis] = np.newaxis
            img = img[slicer]

        nbands = img.shape[bands_axis]
        if nbands not in (1, 3, 4):
            raise NotImplementedError(
                "Can only display images with 1 or 3 bands currently, not {}. "
                "Is axis {} actually your bands axis?".format(nbands, bands_axis)
            )

        if nbands == 4:
            # don't include alpha band in min max
            slicer = [slice(None)] * 3
            slicer[bands_axis] = slice(None, 3)
            spectrals = img[slicer]
        else:
            spectrals = img

        # calculate min and max
        if robust:
            if hasattr(spectrals, "mask"):
                spectrals = spectrals.compressed()  # don't include masked values in percentile
            vmin, vmax = np.percentile(spectrals, 2), np.percentile(spectrals, 98)
            if vmin == vmax:
                robust = False

        if not robust:
            vmin, vmax = np.ma.min(spectrals), np.ma.max(spectrals)

        # matplotlib requires shape (n, m, band)
        disp = np.moveaxis(img, bands_axis, -1).astype(np.float64)

        # rescale
        disp -= vmin
        disp /= (vmax - vmin)
        np.clip(disp, 0, 1, out=disp)  # to coerce away any floating-point errors

        if hasattr(disp, "mask"):
            # turn mask into alpha
            if nbands == 4:
                raise NotImplementedError(
                    "Currently can't supply an image with an explicit alpha band as well as a mask"
                )
            if np.isscalar(disp.mask):
                alpha = np.ones(disp.shape[:2]) * (not disp.mask)
            else:
                alpha = (~disp.mask.any(axis=-1)).astype(disp.dtype)
                if nbands == 1:
                    # to use an alpha channel, matplotlib must have a 4-band image,
                    # so just duplicate the 1 band for r, g, and b
                    disp = np.concatenate([disp] * 3, axis=-1)  # TODO: unnecessary copy of disp's mask
            disp = np.concatenate([disp.data, alpha[:, :, np.newaxis]], axis=-1)

        if disp.shape[-1] == 1:
            # matplotlib takes 1 band images as (n, m), not (n, m, 1)
            disp = disp[:, :, 0]

        ax.imshow(disp, aspect="equal", interpolation=interpolation)
        if title is not None:
            ax.set_title(str(title))
    fig.tight_layout()
    plt.show()



# Copyright 2018 Descartes Labs.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest

from mock import MagicMock
from descarteslabs.client.services.service import Service, JsonApiService


class TestService(unittest.TestCase):
    def test_session_token(self):
        token = "foo.bar.sig"
        service = Service('foo', auth=MagicMock(token=token))
        self.assertEqual(service.session.headers.get('Authorization'), token)


class TestJsonApiService(unittest.TestCase):
    def test_session_token(self):
        token = "foo.bar.sig"
        service = JsonApiService('foo', auth=MagicMock(token=token))
        self.assertEqual(service.session.headers.get('Authorization'), token)


if __name__ == '__main__':
    unittest.main()



# Copyright 2018 Descartes Labs.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
A list-based sequence with helper methods, which serves as the base class for `SceneCollection`.
"""

import itertools
import collections
import six


# TODO: maybe subclass collections.UserList?
class Collection(object):
    """
    List-based sequence with convenience methods for mapping and filtering,
    and NumPy-style fancy indexing
    """
    def __init__(self, iterable=None):
        if iterable is None:
            self._list = []
        else:
            self._list = list(iterable)

    def __getitem__(self, idx):
        """
        self[start:stop:step] <--> Collection(list(self[start:stop:step]))
        self[<list>] <--> Collection(self[i] for i in <list>)
        Can slice like a normal list, or with a list of indices to select
        """
        if isinstance(idx, list) or type(idx).__name__ == 'ndarray':
            subset = [self._list[i] for i in idx]
            return self._cast_and_copy_attrs_to(subset)
        else:
            if isinstance(idx, slice):
                return self._cast_and_copy_attrs_to(self._list[idx])
            else:
                return self._list[idx]

    def __setitem__(self, idx, item):
        """
        Can assign a scalar, or a list of equal length to the slice,
        to any valid slice (including an list of indices)
        """
        if isinstance(idx, (list, slice)) or type(idx).__name__ == 'ndarray':
            if isinstance(idx, slice):
                idx = list(range(*idx.indices(len(self))))
            if isinstance(item, six.string_types) or not isinstance(item, collections.Sequence):  # if scalar
                item = [item] * len(idx)
            item = list(item)
            if len(idx) != len(item):
                raise ValueError("Cannot assign {} items to a slice {} items long".format(len(item), len(idx)))
            for i, x in zip(idx, item):
                self._list[i] = x
        else:
            self._list[idx] = item

    def __iter__(self):
        return iter(self._list)

    def __reversed__(self):
        return reversed(self._list)

    def __contains__(self, other):
        return other in self._list

    def __len__(self):
        return len(self._list)

    def __eq__(self, other):
        return self._list == other

    def __repr__(self):
        return "{}({})".format(self.__class__.__name__, repr(self._list))

    def _cast_and_copy_attrs_to(self, other):
        # used to copy over any attrs a subclass may have set
        other = self.__class__(other)
        for k, v in six.iteritems(self.__dict__):
            if k != "_list":
                setattr(other, k, v)
        return other

    @property
    def each(self):
        """
        Any operations chained onto ``.each`` (attribute access, item access,
        and calls) are applied to each item in the Collection.

        Notes
        -----
            * Add ``.combine()`` at the end of the operations chain
              to combine the results into a list by default, or any
              container type passed into ``.combine()``
            * Use ``.pipe(f, *args, **kwargs)`` to yield
              ``f(x, *args, **kwargs)`` for each item ``x`` yielded by the
              preceeding operations chain

        Examples
        --------
        >>> c = Collection(["one", "two", "three", "four"])
        >>> for x in c.each.capitalize():
        ...     print(x)
        One
        Two
        Three
        Four
        >>> c.each.capitalize()[:2]
        'On'
        'Tw'
        'Th'
        'Fo'
        >>> c.each.capitalize().pipe(len)
        3
        3
        5
        4
        >>> c.each.capitalize().pipe(len).combine(set)
        {3, 4, 5}

        Yields
        ------
            item with all operations following ``.each`` applied to it
        """
        return Eacher(iter(self._list))

    def map(self, f):
        "Returns Collection of ``f`` applied to each item in self"
        res = (f(x) for x in self._list)
        return self._cast_and_copy_attrs_to(res)

    def filter(self, predicate):
        "Returns Collection of items in self for which ``predicate(item)`` is True"
        res = (x for x in self._list if predicate(x))
        return self._cast_and_copy_attrs_to(res)

    def sorted(self, *predicates, **reverse):
        """
        Returns a copy of self, sorted by predicates in ascending order.

        Each predicate can be a key function, or a string of dot-chained attributes
        to use as sort keys. The reverse flag returns results in descending order.

        Examples
        --------
        >>> import collections
        >>> FooBar = collections.namedtuple("FooBar", ["foo", "bar"])
        >>> X = collections.namedtuple("X", "x")
        >>> c = Collection([FooBar(1, X("one")), FooBar(2, X("two")), FooBar(3, X("three"))])

        >>> c.sorted("foo")
        Collection([FooBar(foo=1, bar=X(x='one')), FooBar(foo=2, bar=X(x='two')), FooBar(foo=3, bar=X(x='three'))])
        >>> c.sorted("bar.x")
        Collection([FooBar(foo=1, bar=X(x='one')), FooBar(foo=3, bar=X(x='three')), FooBar(foo=2, bar=X(x='two'))])
        """
        if len(predicates) == 0:
            raise TypeError("No predicate(s) given to sorted")
        predicates = [self._str_to_predicate(p) if isinstance(p, six.string_types) else p for p in predicates]
        if len(predicates) == 1:
            predicate = predicates[0]
        else:
            def predicate(v):
                return tuple(p(v) for p in predicates)

        res = (sorted(self, key=predicate, **reverse))
        return self._cast_and_copy_attrs_to(res)

    def groupby(self, *predicates):
        """
        Groups items by predicates and yields tuple of ``(group, items)``
        for each group, where ``items`` is a Collection.

        Each predicate can be a key function, or a string of dot-chained attributes
        to use as sort keys.

        Examples
        --------
        >>> import collections
        >>> FooBar = collections.namedtuple("FooBar", ["foo", "bar"])
        >>> c = Collection([FooBar("a", True), FooBar("b", False), FooBar("a", False)])

        >>> for group, items in c.groupby("foo"):
        ...     print(group, items)
        a Collection([FooBar(foo='a', bar=True), FooBar(foo='a', bar=False)])
        b Collection([FooBar(foo='b', bar=False)])
        >>> for group, items in c.groupby("bar"):
        ...     print(group, items)
        False Collection([FooBar(foo='b', bar=False), FooBar(foo='a', bar=False)])
        True Collection([FooBar(foo='a', bar=True)])
        """
        if len(predicates) == 0:
            raise TypeError("No predicate(s) given to groupby")
        predicates = [self._str_to_predicate(p) if isinstance(p, six.string_types) else p for p in predicates]
        if len(predicates) == 1:
            predicate = predicates[0]
        else:
            def predicate(v):
                return tuple(p(v) for p in predicates)

        ordered = self.sorted(predicate)
        for group, items in itertools.groupby(ordered, predicate):
            yield group, self._cast_and_copy_attrs_to(items)

    def append(self, x):
        "Append x to end of self"
        self._list.append(x)

    def extend(self, x):
        "Extend self by appending elements from the iterable"
        self._list.extend(x)

    @staticmethod
    def _str_to_predicate(string):
        attrs = string.split(".")

        def predicate(x):
            result = x
            for attr in attrs:
                result = getattr(result, attr)
            return result
        return predicate


class Eacher(object):
    "Applies operations chained onto it to each item in an iterator"
    __slots__ = ("_iterable")

    def __init__(self, iterable):
        self._iterable = iterable

    def __iter__(self):
        return iter(self._iterable)

    def __getattr__(self, attr):
        return Eacher(getattr(x, attr) for x in self)

    def __getitem__(self, idx):
        return Eacher(x[idx] for x in self)

    def __call__(self, *args, **kwargs):
        return Eacher(x(*args, **kwargs) for x in self)

    def combine(self, collection=list):
        "``self.combine(collection) <--> collection(iter(self))``"
        return collection(iter(self))

    def pipe(self, callable, *args, **kwargs):
        "``self.pipe(f, *args, **kwargs) <--> f(x, *args, **kwargs) for x in self``"
        return Eacher(callable(x, *args, **kwargs) for x in self)

    def __repr__(self):
        max_length = 8
        objs = list(itertools.islice(self, max_length))
        try:
            head = [repr(x) for x in objs]
        except Exception:
            return '<%s instance at %#x>' % (self.__class__.__name__, id(self))
        else:
            s = "\n".join(head)
            if len(head) == max_length:
                s = s + "\n..."
            return s



# Copyright 2018 Descartes Labs.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Datasets in the Descartes Labs catalog have many different resolutions and
projections. In two different images, even covering the same place on Earth,
the pixels ``(i, j)`` usually correspond to two different points on the ground.

GeoContexts are a way to ensure multiple images from different sources
are **spatially compatible**---that is, they all have the same shape
(same width and height, in pixels), and the same pixel in each image
corresponds to the same area on Earth.

They do this by simply capturing all the spatial parameters that affect how
imagery is rasterized---namely output resolution, coordinate reference system,
and bounding box---in one object that can be passed into different method calls.
In typical use, these contexts are created for you with reasonable defaults,
so you only need to understand the different parameters when you need more control.

The different subclasses of `GeoContext` implement different
functionality.

* `AOI` clips to arbitrary geometry, and lets you specify any output resolution
  and projection.
* `DLTile` helps you split large regions up into a grid of any spacing and
  resolution, and represents a single tile in that grid, in UTM projection.

Tutorial
========

Often, you don't have to create GeoContexts yourself---an `AOI`
with default parameters is created for you by `scenes.search <scenes._search.search>`
and `Scene.from_id`.

.. ipython:: python

    import descarteslabs as dl

    scene, default_ctx = dl.scenes.Scene.from_id('landsat:LC08:PRE:TOAR:meta_LC80260322013188_v1')
    default_ctx

GeoContexts are immutable; instead, create copies with new values using `AOI.assign`.
(Assigning new values to DLTiles is not yet supported.)

.. ipython:: python

    # let's use a lower resolution to load images faster
    lowres = default_ctx.assign(resolution=75)
    lowres_arr = scene.ndarray("red green blue", lowres)
    @savefig geocontext1.png
    dl.scenes.display(lowres_arr, size=4, title="Default GeoContext, 75-meter resolution")

You can also create GeoContexts explicitly:

.. ipython:: python

    import shapely.affinity

    # make a new polygon half the size of the scene's full extent
    new_cutline = shapely.affinity.scale(scene.geometry, xfact=0.5, yfact=0.5)

    webmerc_cutline_aoi = dl.scenes.AOI(
        geometry=new_cutline,
        resolution=75,
        crs="EPSG:3857"  # "EPSG:3857" is the code for the Web Mercator
    )                    # coordinate reference system, see http://epsg.io/3857

    webmerc_cutline_arr = scene.ndarray("red green blue", webmerc_cutline_aoi)
    @savefig geocontext2.png
    dl.scenes.display(webmerc_cutline_arr, size=4, title="Same scene, with cutline and Web Mercator")

Let's assign our new cutline to the default GeoContext
to see the difference between the coordinate reference systems:

.. ipython:: python

    with_cutline = lowres.assign(geometry=new_cutline)

    with_cutline_arr = scene.ndarray("red green blue", with_cutline)
    @savefig geocontext3.png
    dl.scenes.display(with_cutline_arr, size=4, title="Original GeoContext with new cutline")

Why is there all that empty space around the sides?
We assigned a new geometry, but we didn't change the *bounds*.
Bounds determine the x-y extent that's rasterized; geometry just clips within that.
If shapely is installed, you can pass ``bounds="update"`` to compute new bounds when assinging a new geometry.

.. ipython:: python

    cutline_bounds = lowres.assign(geometry=new_cutline, bounds="update")
    cutline_bounds_arr = scene.ndarray("red green blue", cutline_bounds)
    @savefig geocontext4.png
    dl.scenes.display(cutline_bounds_arr, size=4, title="Original GeoContext, new cutline and bounds")

You can also use DLTiles to split up regions along a grid:

.. ipython:: python

    tiles = dl.scenes.DLTile.from_shape(
        new_cutline, resolution=75, tilesize=256, pad=16
    )
    len(tiles)

    tile0_arr = scene.ndarray("red green blue", tiles[0])
    tile1_arr = scene.ndarray("red green blue", tiles[1])

    @savefig geocontext5.png
    dl.scenes.display(tile0_arr, tile1_arr, title=[tiles[0].key, tiles[1].key], size=3)
"""

import copy

import six
from six.moves import reprlib
from descarteslabs.client.addons import ThirdParty, shapely

from descarteslabs.client.services.raster import Raster

have_shapely = not isinstance(shapely, ThirdParty)


class GeoContext(object):
    """
    Specifies spatial parameters to use when loading a raster
    from the Descartes Labs catalog.

    Two Scenes loaded with the same GeoContext will result in images
    with the same shape (in pixels), covering the same spatial extent,
    regardless of the dimensions or projection of the original data.

    Specifically, a fully-defined GeoContext specifies:

    * geometry to use as a cutline (WGS84), and/or bounds (WGS84)
    * resolution (m)
    * EPSG code of the output coordinate reference system
    * whether to align pixels to the output CRS
      (see docstring for `AOI.align_pixels` for more information)

    GeoContexts are immutable.
    """
    __slots__ = ()

    @property
    def raster_params(self):
        """
        dict: The properties of this GeoContext,
        as keyword arguments to use for `Raster.ndarray` or `Raster.raster`.
        """
        raise NotImplementedError

    def __eq__(self, other):
        """
        Two GeoContexts are equal only if they are the same type,
        and every property is equal.
        """
        if not isinstance(other, self.__class__):
            return False
        for attr in self.__slots__:
            if getattr(self, attr) != getattr(other, attr):
                return False
        return True

    def __repr__(self):
        classname = self.__class__.__name__
        delim = ",\n" + " " * (len(classname) + 1)
        props = delim.join(
            "{}={}".format(attr.lstrip("_"), reprlib.repr(getattr(self, attr)))
            for attr in self.__slots__
        )
        return "{}({})".format(classname, props)


class AOI(GeoContext):
    """
    A GeoContext that clips scenes to a geometry, and/or to square bounds,
    with any output resolution and CRS.

    Examples
    --------

    .. code-block:: python

        cutline_aoi = dl.scenes.AOI(my_geometry, resolution=40)
        aoi_with_cutline_disabled = cutline_aoi.assign(geometry=None)
        no_cutline_aoi = dl.scenes.AOI(geometry=None, resolution=15, bounds=(-40, 35, -39, 36))
        aoi_without_auto_bounds = dl.scenes.AOI(geometry=my_geometry, resolution=15, bounds=(-40, 35, -39, 36))
        aoi_with_specific_pixel_dimensions = dl.scenes.AOI(geometry=my_geometry, shape=(200, 400))
    """
    __slots__ = (
        "_geometry",
        "_resolution",
        "_crs",
        "_align_pixels",
        "_bounds",
        "_shape",
    )

    def __init__(self,
                 geometry=None,
                 resolution=None,
                 crs=None,
                 align_pixels=True,
                 bounds=None,
                 shape=None,
                 ):
        """
        Parameters
        ----------
        geometry: GeoJSON-like dict, object with ``__geo_interface__``; optional
            Clip scenes to this geometry.
            Coordinates must be WGS84 (lat-lon).
            If None, scenes will just be clipped to ``bounds``.
        resolution: float, optional
            Distance, in units of the CRS, that the edge of each pixel
            represents on the ground.
            Can only specify one of ``resolution`` and ``shape``.
        crs: str, optional
            Coordinate Reference System into which scenes will be projected,
            expressed as an EPSG code (like ``"EPSG:4326"``), a PROJ.4 definition,
            or an OGC CRS Well-Known Text string
        align_pixels: bool, optional, default True
            If True, this ensures that, in different Scenes rasterized
            with this same AOI GeoContext, pixels ``(i, j)`` correspond
            to the same area in space. This is accomplished by snapping the
            coordinates of the origin (top-left corner of top-left pixel)
            to a non-fractional interval of ``resolution``.

            If ``align_pixels`` is False, when using scenes with different
            native resolutions and/or projections, pixels at the same indicies
            can be misaligned by a fraction of ``resolution``
            (i.e. correspond to *slighly* different coordinates in space).

            However, this requires warping of the original image, which can be
            undesireable when you want to work with the original data in its
            native resolution and projection.
        bounds: 4-tuple, optional
            Clip scenes to these ``(min_x, min_y, max_x, max_y)`` bounds,
            expressed in WGS84 (lat-lon) coordinates.
            If the Shapely package is installed, ``bounds`` are automatically
            computed from ``geometry`` if not specified.
            Otherwise, ``bounds`` are required.
        shape: 2-tuple, optional
            ``(rows, columns)``, in pixels, the output raster should fit within;
            the longer side of the raster will be min(shape).
            Can only specify one of ``resolution`` and ``shape``.
        """

        self._validate_and_assign(
            geometry,
            resolution,
            crs,
            align_pixels,
            bounds,
            shape,
        )

        # If no bounds were given, use the bounds of the geometry if possible
        if self._bounds is None and self._geometry is not None:
                if have_shapely:
                    self._bounds = self._geometry.bounds
                else:
                    # TODO: _bounds_from_geometry
                    raise NotImplementedError(
                        "Bounds must be set, i.e. to (minx, miny, maxx, maxy) of geometry. "
                        "(Bounds are only inferred from geometry if the shapely package is installed.)"
                    )

    @property
    def geometry(self):
        """
        dict or shapely geometry: Clip scenes to this geometry
        Coordinates must be WGS84 (lat-lon)
        If None, scenes will just be clipped to ``bounds``
        """
        return self._geometry

    @property
    def resolution(self):
        """
        float: Distance, in units of the CRS, that the edge of each pixel
        represents on the ground.
        """
        return self._resolution

    @property
    def crs(self):
        """
        str: Coordinate Reference System into which scenes will be projected,
        expressed as an EPSG code (like ``"EPSG:4326"``), a PROJ.4 definition,
        or an OGC CRS Well-Known Text string
        """
        return self._crs

    @property
    def align_pixels(self):
        """
        bool: If True, this ensures that, in different Scenes rasterized with
        this same AOI GeoContext, pixels ``(i, j)`` correspond to the
        same area in space. This is accomplished by snapping the coordinates of
        the origin (top-left corner of top-left pixel) to a non-fractional
        interval of ``resolution``.

        If ``align_pixels`` is False, when using scenes with different native
        resolutions and/or projections, pixels at the same indicies can be
        misaligned by a fraction of ``resolution`` (i.e. correspond to *slighly*
        different coordinates in space).

        However, this requires warping of the original image, which can be
        undesireable when you want to work with the original data in its native
        resolution and projection.
        """
        return self._align_pixels

    @property
    def bounds(self):
        """
        tuple: Clip scenes to these ``(min_x, min_y, max_x, max_y)`` bounds,
        expressed in WGS84 (lat-lon) coordinates.
        """
        return self._bounds

    @property
    def shape(self):
        """
        tuple: ``(rows, columns)``, in pixels, the output raster should fit within;
        the longer side of the raster will be min(shape).
        """
        return self._shape

    @property
    def raster_params(self):
        """
        dict: The properties of this AOI,
        as keyword arguments to use for ``Raster.ndarray`` or ``Raster.raster``.

        Raises ValueError if ``self.bounds``, ``self.crs``, ``self.resolution``,
        or ``self.align_pixels`` is None, or values are invalid.
        """
        # Ensure that there can be no ambiguity: every parameter must be specified,
        # so every raster call using this context will return spatially equivalent data
        if self._bounds is None:
            raise ValueError("AOI must have bounds specified")
        if self._crs is None:
            raise ValueError("AOI must have CRS specified")
        if self._resolution is None and self._shape is None:
            raise ValueError("AOI must have one of resolution or shape specified")
        if self._align_pixels is None:
            raise ValueError("AOI must have align_pixels specified")

        if have_shapely:
            cutline = shapely.geometry.mapping(self._geometry) if self._geometry is not None else None
        else:
            cutline = self._geometry

        dimensions = (self._shape[1], self._shape[0]) if self._shape is not None else None

        if self._resolution is not None and self._crs == "EPSG:4326":
            # helpful warning about a common mistake
            crs_width = self._bounds[2] - self._bounds[0]
            crs_height = self._bounds[3] - self._bounds[1]
            msg = (
                "Output raster's {dim} ({dim_len:.4f} decimal degrees) is smaller than its resolution "
                "({res:.4f} decimal degrees), meaning it would be less than one pixel {dim_adj}.\n"
                "Remember that resolution is specified in units of the output CRS. "
                "Since your CRS is EPSG:4326 (WGS84 lat-lon), resolution must be given in decimal degrees, not meters."
            )
            if crs_width < self._resolution:
                raise ValueError(msg.format(dim="width", dim_len=crs_width, res=self._resolution, dim_adj="wide"))
            if crs_height < self._resolution:
                raise ValueError(msg.format(dim="height", dim_len=crs_height, res=self._resolution, dim_adj="tall"))

        return {
            "cutline": cutline,
            "resolution": self._resolution,
            "srs": self._crs,
            "bounds_srs": "EPSG:4326",
            "align_pixels": self._align_pixels,
            "bounds": self._bounds,
            "dimensions": dimensions
        }

    @property
    def __geo_interface__(self):
        """
        dict: ``self.geometry`` as a GeoJSON Geometry dict,
        otherwise ``self.bounds`` as a GeoJSON Polygon dict if ``self.geometry`` is None.
        """
        if self._geometry is not None:
            try:
                return self._geometry.__geo_interface__
            except AttributeError:
                return self._geometry
        elif self._bounds is not None:
            return self._polygon_from_bounds(self._bounds)

    def assign(self,
               geometry="unchanged",
               resolution="unchanged",
               crs="unchanged",
               align_pixels="unchanged",
               bounds="unchanged",
               shape="unchanged",
               ):
        """
        Return a copy of the AOI with the given values assigned.

        Note
        ----
            If you are assigning a new geometry and want bounds to updated as
            well, use ``bounds="update"``. (This requires Shapely.)

            If Shapely is not installed, you should usually specify both
            ``geometry`` and ``bounds`` in ``assign``, and ``bounds`` should be
            the ``(min_x, min_y, max_x, max_y)`` of the new geometry.

            Otherwise, if you assign ``geometry`` without changing ``bounds``,
            the new AOI GeoContext will produce rasters with the same
            shape and covering the same spatial area as the old one, just with
            pixels masked out that fall outside your new geometry.

        Returns
        -------
        new : `AOI`
        """
        new = copy.deepcopy(self)
        new._validate_and_assign(
            geometry,
            resolution,
            crs,
            align_pixels,
            bounds,
            shape,
        )
        return new

    def _validate_and_assign(self,
                             geometry,
                             resolution,
                             crs,
                             align_pixels,
                             bounds,
                             shape,
                             ):
        if geometry != "unchanged":
            if have_shapely:
                # convert geometry to shapely
                if geometry is not None and not isinstance(geometry, shapely.geometry.base.BaseGeometry):
                    try:
                        # TODO: help Shapely handle Features and FeatureCollecions
                        geometry = shapely.geometry.shape(geometry)
                    except Exception:
                        raise ValueError(
                            "Could not interpret the given geometry as a Shapely shape. "
                            "Remember that Shapely does not accept GeoJSON Features or FeatureCollecions, "
                            "only geometry objects."
                        )

                if geometry is not None:
                    # test that geometry is in WGS84
                    geom_bounds = geometry.bounds
                    try:
                        self._test_valid_bounds(geom_bounds)
                    except ValueError:
                        six.raise_from(ValueError("Geometry must be in EPSG:4326 (WGS84 lat-lon) coordinates"), None)

                    # TODO: implement _bounds_from_geometry in order to validate CRS without shapely, and because
                    # it'd be helpful anyway (would remove requirement for explicitly setting bounds
                    # if shapely not installed)
            else:  # no shapely
                try:
                    geometry = geometry.__geo_interface__
                except AttributeError:
                    pass
                # TODO: validate geojson if shapely not available

        if bounds != "unchanged":
            # test that bounds are sane, and in WGS84
            if bounds is not None:
                if bounds == "update":
                    if geometry is not None:
                        if have_shapely:
                            bounds = geometry.bounds
                        else:
                            raise NotImplementedError(
                                "Currently, bounds can only be computed from geometry if shapely is installed"
                            )
                    else:
                        raise ValueError("A geometry must be given with which to update the bounds")
                else:
                    self._test_valid_bounds(bounds)
                    bounds = tuple(bounds)

        if shape != "unchanged" and shape is not None:
            if not isinstance(shape, (list, tuple)) or len(shape) != 2:
                raise TypeError("Shape must be a tuple of (rows, columns) in pixels")

        if resolution != "unchanged" and resolution is not None:
            if not isinstance(resolution, (int, float)):
                raise TypeError("Resolution must be an int or float, got type '{}'".format(type(resolution).__name__))
            if resolution <= 0:
                raise ValueError("Resolution must be greater than zero")

        if geometry != "unchanged":
            self._geometry = geometry
        if resolution != "unchanged":
            self._resolution = resolution
        if crs != "unchanged":
            self._crs = crs
        if align_pixels != "unchanged":
            self._align_pixels = align_pixels
        if bounds != "unchanged":
            self._bounds = bounds
        if shape != "unchanged":
            self._shape = shape

        if self._resolution is not None and self._shape is not None:
            raise ValueError("Cannot set both resolution and shape")

        # test that bounds and geometry actually intersect
        # TODO: do this without shapely
        if have_shapely and self._geometry is not None and self._bounds is not None:
            bounds_shp = shapely.geometry.box(*self._bounds)
            if not bounds_shp.intersects(self._geometry):
                raise ValueError(
                    "Geometry and bounds do not intersect. This would result in all data being masked. "
                    "If you're assigning new geometry, assign new bounds as well "
                    "(use `bounds='update'` to use the bounds of the new geometry)."
                )

    @classmethod
    def _polygon_from_bounds(cls, bounds):
        return {
            "type": "Polygon",
            "coordinates": ((
                (bounds[2], bounds[1]),
                (bounds[2], bounds[3]),
                (bounds[0], bounds[3]),
                (bounds[0], bounds[1]),
                (bounds[2], bounds[1]),
            ),)
        }

    @classmethod
    def _test_valid_bounds(cls, bounds):
        """
        Test given bounds are in correct order and are valid WGS84 lat-lon values

        Raises TypeError or ValueError if bounds are invalid
        """
        try:
            if not isinstance(bounds, (list, tuple)):
                raise TypeError()

            if len(bounds) != 4:
                raise ValueError(
                    "Bounds must a sequence of (minx, miny, maxx, maxy), "
                    "got sequence of length {}".format(len(bounds))
                )
        except TypeError:
            six.raise_from(
                TypeError("Bounds must a sequence of (minx, miny, maxx, maxy), got {}".format(type(bounds))),
                None
            )
        if not (-180 < bounds[0] < 180 and
                -90 < bounds[1] < 90 and
                -180 < bounds[2] < 180 and
                -90 < bounds[3] < 90):
            raise ValueError("Bounds must be in EPSG:4326 (WGS84 lat-lon) coordinates")

        if bounds[0] >= bounds[2]:
            raise ValueError("minx >= maxx in given bounds, should be (minx, miny, maxx, maxy)")
        if bounds[1] >= bounds[3]:
            raise ValueError("miny >= maxy in given bounds, should be (minx, miny, maxx, maxy)")


class DLTile(GeoContext):
    """
    A GeoContext that clips and projects Scenes to a single DLTile.

    DLTiles allow you to define a grid of arbitrary spacing, resolution,
    and overlap that can cover the globe.
    DLTiles are always in a UTM projection.
    """
    __slots__ = (
        "_key",
        "_resolution",
        "_tilesize",
        "_pad",
        "_crs",
        "_bounds",
        "_geometry",
        "_zone",
        "_ti",
        "_tj",
    )

    def __init__(self, dltile_dict):
        """
        ``__init__`` instantiates a DLTile from a dict returned by `Raster.dltile`.

        It's preferred to use the `DLTile.from_latlon`, `DLTile.from_shape`,
        or `DLTile.from_key` class methods to construct a DLTile GeoContext.
        """
        self._geometry = shapely.geometry.shape(dltile_dict['geometry']) if have_shapely else dltile_dict['geometry']
        properties = dltile_dict['properties']
        self._key = properties["key"]
        self._resolution = properties["resolution"]
        self._tilesize = properties["tilesize"]
        self._pad = properties["pad"]
        self._crs = properties["cs_code"]
        self._bounds = tuple(properties["outputBounds"])
        self._zone = properties["zone"]
        self._ti = properties["ti"]
        self._tj = properties["tj"]

    @classmethod
    def from_latlon(cls, lat, lon, resolution, tilesize, pad, raster_client=None):
        """
        Return a DLTile GeoContext that covers a latitude/longitude

        Where the point falls within the tile will vary, depending on the point
        and tiling parameters.

        Parameters
        ----------
        lat : float
            Latitude (WGS84)
        lon : float
            Longitude (WGS84)
        resolution : float
            Distance, in meters, that the edge of each pixel represents on the ground
        tilesize : int
            Length of each side of the tile, in pixels
        pad : int
            Number of extra pixels by which each side of the tile is buffered.
            This determines the number of pixels by which two tiles overlap.
        raster_client : descarteslabs.client.services.Raster, optional, default None
            Unneeded in general use; lets you use a specific client instance
            with non-default auth and parameters.

        Returns
        -------
        tile : DLTile
        """
        if raster_client is None:
            raster_client = Raster()
        tile = raster_client.dltile_from_latlon(lat, lon, resolution, tilesize, pad)
        return cls(tile)

    @classmethod
    def from_shape(cls, shape, resolution, tilesize, pad, raster_client=None):
        # TODO : non-overlapping tiles across UTM zones
        """
        Return a list of DLTiles that intersect the given geometry

        Parameters
        ----------
        shape : GeoJSON-like
            A GeoJSON dict, or object with a __geo_interface__. Must be in
            EPSG:4326 (WGS84 lat-lon) projection.
        resolution : float
            Distance, in meters, that the edge of each pixel represents on the ground
        tilesize : int
            Length of each side of the tile, in pixels
        pad : int
            Number of extra pixels by which each side of the tile is buffered.
            This determines the number of pixels by which two tiles overlap.
        raster_client : descarteslabs.client.services.Raster, optional, default None
            Unneeded in general use; lets you use a specific client instance
            with non-default auth and parameters.

        Returns
        -------
        tiles : List[DLTile]
        """
        if raster_client is None:
            raster_client = Raster()

        if hasattr(shape, "__geo_interface__"):
            shape = shape.__geo_interface__

        tiles_fc = raster_client.dltiles_from_shape(resolution=resolution, tilesize=tilesize, pad=pad, shape=shape)
        return [cls(tile) for tile in tiles_fc["features"]]

    @classmethod
    def from_key(cls, dltile_key, raster_client=None):
        """
        Return a DLTile GeoContext from a DLTile key.

        Parameters
        ----------
        dltile_key : str
            DLTile key, e.g. '128:16:960.0:15:-1:37'
        raster_client : descarteslabs.client.services.Raster, optional, default None
            Unneeded in general use; lets you use a specific client instance
            with non-default auth and parameters.

        Returns
        -------
        tile: DLTile
        """
        if raster_client is None:
            raster_client = Raster()
        tile = raster_client.dltile(dltile_key)
        return cls(tile)

    @property
    def key(self):
        """
        str: The DLTile's key, which encodes the tiling parameters,
        and which number in the grid this tile is.
        """
        return self._key

    @property
    def resolution(self):
        "float: Distance, in meters, that the edge of each pixel represents on the ground"
        return self._resolution

    @property
    def tilesize(self):
        """
        int: Length of each side of the tile, in pixels.
        Note that the total number of pixels along each side of an image is
        ``tile_size + 2*padding``
        """
        return self._tilesize

    @property
    def pad(self):
        """
        int: Number of extra pixels by which each side of the tile is buffered.
        This determines the number of pixels by which two tiles overlap.
        """
        return self._pad

    @property
    def crs(self):
        """
        str: Coordinate Reference System into which scenes will be projected.
        For DLTiles, this is always a UTM projection, given as an EPSG code.
        """
        return self._crs

    @property
    def bounds(self):
        """
        tuple: The ``(min_x, min_y, max_x, max_y)`` of the area covered by
        this DLTile, in UTM coordinates
        """
        # QUESTION: should this be in WGS84 to be consistent with the rest of GeoContext
        return self._bounds

    @property
    def geometry(self):
        """
        shapely.geometry.Polygon, or dict: The polygon covered by this DLTile
        in WGS84 (lat-lon) coordinates
        """
        return self._geometry

    @property
    def zone(self):
        "int: The UTM zone of this tile"
        return self._zone

    @property
    def ti(self):
        "int: The y-index of this tile in its grid"
        return self._ti

    @property
    def tj(self):
        "int: The x-index of this tile in its grid"
        return self._tj

    @property
    def raster_params(self):
        """
        dict: The properties of this DLTile,
        as keyword arguments to use for `Raster.ndarray` or `Raster.raster`.
        """
        return {
            "dltile": self._key,
            "align_pixels": False
            # QUESTION: shouldn't align_pixels be True?
            # based on the GDAL documentation for `-tap`, seems like that should be true
            # to ensure that pixels of images with different resolutions/projections
            # are aligned with the same dltile. otherwise, pixel (0,0) in 1 image could be at
            # different coordinates than the other
        }

    @property
    def __geo_interface__(self):
        "dict: ``self.geometry`` as a GeoJSON dict"
        try:
            return self._geometry.__geo_interface__
        except AttributeError:
            return self._geometry

# TODO: XYZTile?



# Copyright 2018 Descartes Labs.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Create a SceneCollection by searching:

.. ipython::

    In [1]: import descarteslabs as dl

    In [2]: import numpy as np

    In [3]: aoi_geometry = {'type': 'Polygon',
       ...:  'coordinates': (((-95.27841503861751, 42.76556057019057),
       ...:    (-93.15675252485482, 42.36289849433184),
       ...:    (-93.73350276458868, 40.73810018004927),
       ...:    (-95.79766011799035, 41.13809376845988),
       ...:    (-95.27841503861751, 42.76556057019057)),)}

    In [4]: scenes, ctx = dl.scenes.search(aoi_geometry, products=["landsat:LC08:PRE:TOAR"], limit=10)

    @doctest
    In [5]: scenes
    Out[5]:
    SceneCollection of 10 scenes
      * Dates: Apr 18, 2013 to Sep 09, 2013
      * Products: landsat:LC08:PRE:TOAR: 10

Use `SceneCollection.each` and `SceneCollection.filter` to subselect Scenes you want:

.. ipython::

    In [6]: # which month is each scene from?

    @doctest
    In [7]: scenes.each.properties.date.month.combine()
    Out[7]: [4, 5, 5, 6, 6, 7, 7, 8, 8, 9]

    In [8]: spring_scenes = scenes.filter(lambda s: s.properties.date.month <= 6)

    @doctest
    In [9]: spring_scenes
    Out[9]:
    SceneCollection of 5 scenes
      * Dates: Apr 18, 2013 to Jun 21, 2013
      * Products: landsat:LC08:PRE:TOAR: 5

Operate on related Scenes with `SceneCollection.groupby`:

.. ipython::

    @doctest
    In [14]: for month, month_scenes in spring_scenes.groupby("properties.date.month"):
       ....:    print("Month {}: {} scenes".format(month, len(month_scenes)))
       ....:
    Month 4: 1 scenes
    Month 5: 2 scenes
    Month 6: 2 scenes

Load data with `SceneCollection.stack` or `SceneCollection.mosaic`:

.. ipython::

    In [10]: ctx_lowres = ctx.assign(resolution=120)

    In [11]: stack = spring_scenes.stack("red green blue", ctx_lowres)

    In [12]: stack.shape
    Out[12]: (5, 3, 3690, 3724)
"""

from __future__ import division
import collections
import six
import logging
import json

from descarteslabs.client.addons import ThirdParty, concurrent, shapely, numpy as np

from descarteslabs.client.services.raster import Raster
from descarteslabs.client.exceptions import NotFoundError, BadRequestError

from .collection import Collection
from .scene import Scene

have_shapely = not isinstance(shapely, ThirdParty)


class SceneCollection(Collection):
    """
    Holds Scenes, with methods for loading their data.

    As a subclass of `Collection`, the `filter`, `map`, and `groupby`
    methods and `each` property simplify inspection and subselection of
    contianed Scenes.

    `stack` and `mosaic` rasterize all contained Scenes into an ndarray
    using the a GeoContext.
    """
    def __init__(self, iterable=None, raster_client=None):
        super(SceneCollection, self).__init__(iterable)
        self._raster_client = raster_client if raster_client is not None else Raster()

    def map(self, f):
        """
        Returns list of ``f`` applied to each item in self,
        or SceneCollection if ``f`` returns Scenes
        """
        res = super(SceneCollection, self).map(f)
        if all(isinstance(x, Scene) for x in res):
            return res
        else:
            return list(res)

    def stack(self,
              bands,
              ctx,
              flatten=None,
              mask_nodata=True,
              mask_alpha=True,
              bands_axis=1,
              raster_info=False,
              max_workers=None,
              ):
        """
        Load bands from all scenes and stack them into a 4D ndarray,
        optionally masking invalid data.

        Parameters
        ----------
        bands : str or Sequence[str]
            Band names to load. Can be a single string of band names
            separated by spaces (``"red green blue"``),
            or a sequence of band names (``["red", "green", "blue"]``).
            If the alpha band is requested, it must be last in the list
            to reduce rasterization errors.
        ctx : GeoContext
            A GeoContext to use when loading each Scene
        flatten : str, Sequence[str], callable, or Sequence[callable], default None
            "Flatten" groups of Scenes in the stack into a single layer by mosaicking
            each group (such as Scenes from the same day), then stacking the mosaics.

            ``flatten`` takes the same predicates as `Collection.groupby`, such as
            ``"properties.date"`` to mosaic Scenes acquired at the exact same timestamp,
            or ``["properties.date.year", "properties.date.month", "properties.date.day"]``
            to combine Scenes captured on the same day (but not necessarily the same time).

            This is especially useful when ``ctx`` straddles a scene boundary
            and contains one image captured right after another. Instead of having
            each as a separate layer in the stack, you might want them combined.

            Note that indicies in the returned ndarray will no longer correspond to
            indicies in this SceneCollection, since multiple Scenes may be combined into
            one layer in the stack. You can call ``groupby`` on this SceneCollection
            with the same parameters to iterate through groups of Scenes in equivalent
            order to the returned ndarray.

            Additionally, the order of scenes in the ndarray will change:
            they'll be sorted by the parameters to ``flatten``.
        mask_nodata : bool, default True
            Whether to mask out values in each band of each scene that equal
            that band's ``nodata`` sentinel value.
        mask_alpha : bool, default True
            Whether to mask pixels in all bands of each scene where
            the alpha band is 0.
        bands_axis : int, default 1
            Axis along which bands should be located.
            If 1, the array will have shape ``(scene, band, y, x)``, if -1,
            it will have shape ``(scene, y, x, band)``, etc.
            A bands_axis of 0 is currently unsupported.
        raster_info : bool, default False
            Whether to also return a list of dicts about the rasterization of
            each scene, including the coordinate system WKT
            and geotransform matrix.
            Generally only useful if you plan to upload data derived from this
            scene back to the Descartes catalog, or use it with GDAL.
        max_workers : int, default None
            Maximum number of threads to use to parallelize individual ndarray
            calls to each Scene.
            If None, it defaults to the number of processors on the machine,
            multiplied by 5.
            Note that unnecessary threads *won't* be created if ``max_workers``
            is greater than the number of Scenes in the SceneCollection.

        Returns
        -------
        arr : ndarray
            Returned array's shape is ``(scene, band, y, x)`` if bands_axis is 1,
            or ``(scene, y, x, band)`` if bands_axis is -1.
            If ``mask_nodata`` or ``mask_alpha`` is True, arr will be a masked array.
        raster_info : List[dict]
            If ``raster_info=True``, a list of raster information dicts for each scene
            is also returned

        Raises
        ------
        ValueError
            If requested bands are unavailable, or band names are not given
            or are invalid.
            If not all required parameters are specified in the GeoContext.
            If the SceneCollection is empty.
        NotFoundError
            If a Scene's id cannot be found in the Descartes Labs catalog
        BadRequestError
            If the Descartes Labs platform is given unrecognized parameters
        """
        if len(self) == 0:
            raise ValueError("This SceneCollection is empty")

        kwargs = dict(
            mask_nodata=mask_nodata,
            mask_alpha=mask_alpha,
            bands_axis=bands_axis,
            raster_info=raster_info,
        )

        if bands_axis == 0 or bands_axis == -4:
            raise NotImplementedError(
                "bands_axis of 0 is currently unsupported for `SceneCollection.stack`. "
                "If you require this shape, try ``np.moveaxis(my_stack, 1, 0)`` on the returned ndarray."
            )
        elif bands_axis > 0:
            kwargs['bands_axis'] = bands_axis - 1  # the bands axis for each component ndarray call in the stack

        if flatten is not None:
            if isinstance(flatten, six.string_types) or not hasattr(flatten, "__len__"):
                flatten = [flatten]
            scenes = [sc if len(sc) > 1 else sc[0] for group, sc in self.groupby(*flatten)]
        else:
            scenes = self

        full_stack = None
        mask = None
        if raster_info:
            raster_infos = [None] * len(scenes)

        bands = Scene._bands_to_list(bands)
        pop_alpha = False
        if (mask_nodata or mask_alpha) and "alpha" not in bands:
            pop_alpha = True
            bands.append("alpha")
        # Pre-check that all bands and alpha are available in all Scenes, and all have the same dtypes
        self._common_data_type(bands)
        if pop_alpha:
            bands.pop(-1)

        def threaded_ndarrays():
            def data_loader(scene_or_scenecollection, bands, ctx, **kwargs):
                ndarray_kwargs = dict(kwargs, raster_client=self._raster_client)
                if isinstance(scene_or_scenecollection, self.__class__):
                    return lambda: scene_or_scenecollection.mosaic(bands, ctx, **kwargs)
                else:
                    return lambda: scene_or_scenecollection.ndarray(bands, ctx, **ndarray_kwargs)

            try:
                futures = concurrent.futures
            except ImportError:
                logging.warning(
                    "Failed to import concurrent.futures. ndarray calls will be serial."
                )
                for i, scene_or_scenecollection in enumerate(scenes):
                    yield i, data_loader(scene_or_scenecollection, bands, ctx, **kwargs)()
            else:
                with futures.ThreadPoolExecutor(max_workers=max_workers) as executor:
                    future_ndarrays = {}
                    for i, scene_or_scenecollection in enumerate(scenes):
                        future_ndarray = executor.submit(data_loader(scene_or_scenecollection, bands, ctx, **kwargs))
                        future_ndarrays[future_ndarray] = i
                    for future in futures.as_completed(future_ndarrays):
                        i = future_ndarrays[future]
                        result = future.result()
                        yield i, result

        for i, arr in threaded_ndarrays():
            if raster_info:
                arr, raster_meta = arr
                raster_infos[i] = raster_meta

            if full_stack is None:
                stack_shape = (len(scenes),) + arr.shape
                full_stack = np.empty(stack_shape, dtype=arr.dtype)
                if isinstance(arr, np.ma.MaskedArray):
                    mask = np.empty(stack_shape, dtype=bool)

            if isinstance(arr, np.ma.MaskedArray):
                full_stack[i] = arr.data
                mask[i] = arr.mask
            else:
                full_stack[i] = arr

        if mask is not None:
            full_stack = np.ma.MaskedArray(full_stack, mask, copy=False)
        if raster_info:
            return full_stack, raster_infos
        else:
            return full_stack

    def mosaic(self,
               bands,
               ctx,
               mask_nodata=True,
               mask_alpha=True,
               bands_axis=0,
               raster_info=False,
               ):
        """
        Load bands from all scenes, combining them into a single 3D ndarray
        and optionally masking invalid data.

        Where multiple scenes overlap, only data from the scene that comes last
        in the SceneCollection is used.

        Parameters
        ----------
        bands : str or Sequence[str]
            Band names to load. Can be a single string of band names
            separated by spaces (``"red green blue"``),
            or a sequence of band names (``["red", "green", "blue"]``).
            If the alpha band is requested, it must be last in the list
            to reduce rasterization errors.
        ctx : GeoContext
            A GeoContext to use when loading each Scene
        mask_nodata : bool, default True
            Whether to mask out values in each band that equal
            that band's ``nodata`` sentinel value.
        mask_alpha : bool, default True
            Whether to mask pixels in all bands where the alpha band of all scenes is 0.
        bands_axis : int, default 0
            Axis along which bands should be located in the returned array.
            If 0, the array will have shape ``(band, y, x)``,
            if -1, it will have shape ``(y, x, band)``.

            It's usually easier to work with bands as the outermost axis,
            but when working with large arrays, or with many arrays concatenated
            together, NumPy operations aggregating each xy point across bands
            can be slightly faster with bands as the innermost axis.
        raster_info : bool, default False
            Whether to also return a dict of information about the rasterization
            of the scenes, including the coordinate system WKT and geotransform matrix.
            Generally only useful if you plan to upload data derived
            from this scene back to the Descartes catalog, or use it with GDAL.

        Returns
        -------
        arr : ndarray
            Returned array's shape will be ``(band, y, x)`` if ``bands_axis``
            is 0, and ``(y, x, band)`` if ``bands_axis`` is -1.
            If ``mask_nodata`` or ``mask_alpha`` is True, arr will be a masked array.
        raster_info : dict
            If ``raster_info=True``, a raster information dict is also returned.

        Raises
        ------
        ValueError
            If requested bands are unavailable, or band names are not given
            or are invalid.
            If not all required parameters are specified in the GeoContext.
            If the SceneCollection is empty.
        NotFoundError
            If a Scene's ID cannot be found in the Descartes Labs catalog
        BadRequestError
            If the Descartes Labs platform is given unrecognized parameters
        """
        if len(self) == 0:
            raise ValueError("This SceneCollection is empty")

        if not (-3 < bands_axis < 3):
            raise ValueError("Invalid bands_axis; axis {} would not exist in a 3D array".format(bands_axis))

        bands = Scene._bands_to_list(bands)

        if mask_alpha:
            try:
                alpha_i = bands.index("alpha")
            except ValueError:
                bands.append("alpha")
                drop_alpha = True
            else:
                if alpha_i != len(bands) - 1:
                    raise ValueError("Alpha must be the last band in order to reduce rasterization errors")
                drop_alpha = False

        common_data_type = self._common_data_type(bands)

        raster_params = ctx.raster_params
        full_raster_args = dict(
            inputs=[scene.properties["id"] for scene in self],
            order="gdal",
            bands=bands,
            scales=None,
            data_type=common_data_type,
            **raster_params
        )

        try:
            arr, info = self._raster_client.ndarray(**full_raster_args)
        except NotFoundError as e:
            raise NotFoundError(
                "Some or all of these IDs don't exist in the Descartes catalog: {}".format(full_raster_args["inputs"])
            )
        except BadRequestError as e:
            msg = ("Error with request:\n"
                   "{err}\n"
                   "For reference, dl.Raster.ndarray was called with these arguments:\n"
                   "{args}")
            msg = msg.format(err=e, args=json.dumps(full_raster_args, indent=2))
            six.raise_from(BadRequestError(msg), None)

        if len(arr.shape) == 2:
            # if only 1 band requested, still return a 3d array
            arr = arr[np.newaxis]

        if mask_nodata or mask_alpha:
            if mask_alpha:
                alpha = arr[-1]
                if drop_alpha:
                    arr = arr[:-1]
                    bands.pop(-1)

            mask = np.zeros_like(arr, dtype=bool)

            if mask_nodata:
                # collect all possible nodata values per band,
                # in case different products have different nodata values for the same-named band
                # QUESTION: is this overkill?
                band_nodata_values = collections.defaultdict(set)
                for scene in self:
                    scene_bands = scene.properties["bands"]
                    for bandname in bands:
                        band_nodata_values[bandname].add(scene_bands[bandname].get('nodata'))

                for i, bandname in enumerate(bands):
                    for nodata in band_nodata_values[bandname]:
                        if nodata is not None:
                            mask[i] |= arr[i] == nodata

            if mask_alpha:
                mask |= alpha == 0

            arr = np.ma.MaskedArray(arr, mask, copy=False)

        if bands_axis != 0:
            arr = np.moveaxis(arr, 0, bands_axis)
        if raster_info:
            return arr, info
        else:
            return arr

    def __repr__(self):
        parts = ["SceneCollection of {} scene{}".format(len(self), "" if len(self) == 1 else "s")]
        try:
            first = min(self.each.properties.date)
            last = max(self.each.properties.date)
            dates = "  * Dates: {:%b %d, %Y} to {:%b %d, %Y}".format(first, last)
            parts.append(dates)
        except Exception:
            pass

        try:
            products = self.each.properties.product.combine(collections.Counter)
            if len(products) > 0:
                products = ", ".join("{}: {}".format(k, v) for k, v in six.iteritems(products))
                products = "  * Products: {}".format(products)
                parts.append(products)
        except Exception:
            pass

        return "\n".join(parts)

    def _common_data_type(self, bands):
        data_types = [scene._common_data_type_of_bands(bands) for scene in self]
        common_data_type = None
        for i, data_type in enumerate(data_types):
            if common_data_type is None:
                common_data_type = data_type
            else:
                if data_type != common_data_type:
                    raise ValueError(
                        "Bands must all have the same dtype in every Scene. "
                        "The requested bands in Scene {} have dtype '{}', but all prior Scenes had dtype '{}'"
                        .format(i, data_type, common_data_type)
                    )
        return common_data_type



# Copyright 2018 Descartes Labs.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# flake8: noqa
from .client.auth import Auth
from .client import services
from .client.services import *

try:
    from . import scenes
except ImportError:
    pass

descartes_auth = Auth.from_environment_or_token_json()
metadata = Metadata(auth=descartes_auth)
places = Places(auth=descartes_auth)
raster = Raster(auth=descartes_auth)

__all__ = ["descartes_auth", "metadata", "places", "raster", "Auth"] + services.__all__



# Copyright 2018 Descartes Labs.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
The Scene class holds metadata about a single scene in the Descartes Labs catalog.

Example
-------
>>> import descarteslabs as dl
>>> scene, ctx = dl.scenes.Scene.from_id("landsat:LC08:PRE:TOAR:meta_LC80270312016188_v1")
>>> ctx  # a default GeoContext to use when loading raster data from this Scene
AOI(geometry=None,
    resolution=15,
    crs='EPSG:32615',
    align_pixels=True,
    bounds=(-95.8364984, 40.703737, -93.1167728, 42.7999878),
    shape=None)
>>> scene.properties.id
'landsat:LC08:PRE:TOAR:meta_LC80270312016188_v1'
>>> scene.properties.date
datetime.datetime(2016, 7, 6, 16, 59, 42, 753476)
>>> scene.properties.bands.red.resolution
15
>>> arr = scene.ndarray("red green blue", ctx)
>>> type(arr)
<class 'numpy.ma.core.MaskedArray'>
>>> arr.shape
(3, 15259, 15340)
"""

from __future__ import division
import six
import json
import datetime

from descarteslabs.client.addons import ThirdParty, shapely, numpy as np

from descarteslabs.client.services.raster import Raster
from descarteslabs.client.services.metadata import Metadata
from descarteslabs.client.exceptions import NotFoundError, BadRequestError
from descarteslabs.common.dotdict import DotDict

from . import geocontext


def _strptime_helper(s):
    formats = [
        '%Y-%m-%dT%H:%M:%S.%fZ',
        '%Y-%m-%dT%H:%M:%SZ',
        '%Y-%m-%dT%H:%M:%S.%f+00:00',
        '%Y-%m-%dT%H:%M:%S+00:00',
        '%Y-%m-%dT%H:%M:%S'
    ]

    for fmt in formats:
        try:
            return datetime.datetime.strptime(s, fmt)
        except ValueError:
            pass

    return None


class Scene(object):
    """
    Object holding metadata about a single scene in the Descartes Labs catalog.

    A Scene is structured like a GeoJSON Feature, with geometry and properties.

    Attributes
    ----------
    geometry : shapely.geometry.Polygon or dict
        The region the scene's data covers, in WGS84 (lat-lon) coordinates.
        If the Shapely package is installed, it will be a shapely Polygon,
        otherwise a dict of a GeoJSON Polygon
    properties : DotDict
        Metadata about the scene. Some fields will vary between products,
        but these will be present:

        * ``id`` : str
            Descartes Labs ID of this scene
        * ``crs`` : str
            Native coordinate reference system of the scene,
            as an EPSG code or PROJ.4 definition
        * ``date`` : datetime.datetime
            ``'acquired'`` date parsed as a Python datetime if set,
            otherwise None
        * ``bands`` : DotDict[str, DotDict]
            Metadata about the bands available in the scene,
            as the mapping ``{band name: band metadata}``.

            Band names are either the band's ``name`` field (like "red"),
            or for derived bands, the band's ``id`` (like "derived:ndvi").

            Each band metadata dict should contain these fields:

            * ``id`` : str
                Descartes Labs ID of the band;
                unique to every band of every product
            * ``name`` : str
                Human-readable name of the band
            * ``dtype`` : str
                Native type in which the band's data is stored
            * ``data_range`` : list
                List of [min, max] values the band's data can have

            These fields are useful and available in most products,
            but may not always be available:

            * ``resolution`` : float
                Native resolution of the band, in ``resolution_unit``,
                that the edge of each pixel represents on the ground
            * ``resolution_unit`` : str
                Units of ``resolution`` field, such as ``"m"``
            * ``physical_range`` : list
                [min, max] range of values the band's data *represents*.
                Values of data have physical meaning
                (such as a reflectance fraction from 0-1), but often
                those values are remapped to a different numerical range
                for more efficient storage (since fixed-point integers require
                less space than floats). To return data to numbers
                with physical meaning, they should be mapped
                from ``data_range`` to ``physical_range``.
            * ``wavelength_min``
                Minimum wavelength captured by the sensor in this band
            * ``wavelength_center``
                Central wavelength captured by the sensor in this band
            * ``wavelength_max``
                Maximum wavelength captured by the sensor in this band
            * ``wavelength_unit``
                Units of the wavelength fields, such as ``"nm"``
    """
    def __init__(self, scene_dict, bands_dict):
        """
        ``__init__`` instantiates a Scene from a dict returned by `Metadata.search`
        and `Metadata.get_bands_by_id`.

        It's preferred to use `Scene.from_id` or `scenes.search <scenes._search.search>` instead.
        """
        self.geometry = (shapely.geometry.shape(scene_dict["geometry"])
                         if not isinstance(shapely, ThirdParty)
                         else scene_dict["geometry"])
        properties = scene_dict["properties"]
        properties["id"] = scene_dict["id"]
        properties["bands"] = self._scenes_bands_dict(bands_dict)
        properties["crs"] = (properties.pop("cs_code")
                             if "cs_code" in properties
                             else properties.get("proj4"))

        if 'acquired' in properties:
            properties["date"] = _strptime_helper(properties["acquired"])
        else:
            properties["date"] = None

        self.properties = properties

    @classmethod
    def from_id(cls, scene_id, metadata_client=None):
        """
        Return the metadata for a Descartes Labs scene ID as a Scene object.

        Also returns a GeoContext with reasonable defaults to use
        when loading the Scene's ndarray.

        Parameters
        ----------
        scene_id: str
            Descartes Labs scene ID,
            e.g. "landsat:LC08:PRE:TOAR:meta_LC80270312016188_v1"
        metadata_client : Metadata, optional
            Unneeded in general use; lets you use a specific client instance
            with non-default auth and parameters.

        Returns
        -------
        scene: Scene
            Scene instance with metadata loaded from the Descartes Labs catalog
        ctx: AOI
            A default GeoContext useful for loading this Scene.
            These defaults are used:

            * bounds: bounds of the Scene's geometry
            * resolution: the finest resolution of any band in the scene
            * crs: native CRS of the Scene (generally, a UTM CRS)
            * align_pixels: True

        Example
        -------
        >>> import descarteslabs as dl
        >>> scene, ctx = dl.scenes.Scene.from_id("landsat:LC08:PRE:TOAR:meta_LC80260322016197_v1")
        >>> ctx
        AOI(geometry=None,
            resolution=15,
            crs='EPSG:32615',
            align_pixels=True,
            bounds=(-94.724166, 39.2784859, -92.0686956, 41.3717716),
            shape=None)
        >>> scene.properties.date
        datetime.datetime(2016, 7, 15, 16, 53, 59, 495435)

        Raises
        ------
        NotFoundError
            If the ``scene_id`` cannot be found in the Descartes Labs catalog
        """
        if metadata_client is None:
            metadata_client = Metadata()

        metadata = metadata_client.get(scene_id)
        metadata = {
            "type": "Feature",
            "geometry": metadata.pop("geometry"),
            "id": metadata.pop("id"),
            "key": metadata.pop("key"),
            "properties": metadata
        }

        bands = metadata_client.get_bands_by_id(scene_id)
        scene = cls(metadata, bands)

        # TODO: not sure what default res should be
        try:
            default_resolution = min(filter(None, [b.get("resolution") for b in six.itervalues(bands)]))
        except ValueError:
            default_resolution = 100

        # QUESTION: default bounds will now be in WGS84, not UTM
        # indeed, there's no way to give bounds in UTM besides with a DLTile,
        # which means you could get off-by-one issues with loading an entire scene
        # at native resolution, where the WGS84 bounds result in a slightly differently
        # sized raster than native UTM bounds would with reprojection errors
        try:
            bounds = scene.geometry.bounds
        except AttributeError:
            xs, ys = zip(*scene.geometry["coordinates"][0])
            bounds = (min(xs), min(ys), max(xs), max(ys))

        default_ctx = geocontext.AOI(bounds=bounds,
                                     resolution=default_resolution,
                                     crs=scene.properties["crs"]
                                     )

        return scene, default_ctx

    def ndarray(self,
                bands,
                ctx,
                mask_nodata=True,
                mask_alpha=True,
                bands_axis=0,
                raster_info=False,
                raster_client=None
                ):
        """
        Load bands from this scene as an ndarray, optionally masking invalid data.

        Parameters
        ----------
        bands : str or Sequence[str]
            Band names to load. Can be a single string of band names
            separated by spaces (``"red green blue derived:ndvi"``),
            or a sequence of band names (``["red", "green", "blue", "derived:ndvi"]``).
            Names must be keys in ``self.properties.bands``.
            If the alpha band is requested, it must be last in the list
            to reduce rasterization errors.
        ctx : GeoContext
            A GeoContext to use when loading this Scene
        mask_nodata : bool, default True
            Whether to mask out values in each band that equal
            that band's ``nodata`` sentinel value.
        mask_alpha : bool, default True
            Whether to mask pixels in all bands where the alpha band is 0.
        bands_axis : int, default 0
            Axis along which bands should be located in the returned array.
            If 0, the array will have shape ``(band, y, x)``, if -1,
            it will have shape ``(y, x, band)``.

            It's usually easier to work with bands as the outermost axis,
            but when working with large arrays, or with many arrays concatenated
            together, NumPy operations aggregating each xy point across bands
            can be slightly faster with bands as the innermost axis.
        raster_info : bool, default False
            Whether to also return a dict of information about the rasterization
            of the scene, including the coordinate system WKT and geotransform matrix.
            Generally only useful if you plan to upload data derived
            from this scene back to the Descartes catalog, or use it with GDAL.
        raster_client : Raster, optional
            Unneeded in general use; lets you use a specific client instance
            with non-default auth and parameters.

        Returns
        -------
        arr : ndarray
            Returned array's shape will be ``(band, y, x)`` if bands_axis is 0,
            ``(y, x, band)`` if bands_axis is -1
            If ``mask_nodata`` or ``mask_alpha`` is True, arr will be a masked array.
        raster_info : dict
            If ``raster_info=True``, a raster information dict is also returned.

        Example
        -------
        >>> import descarteslabs as dl
        >>> scene, ctx = dl.scenes.Scene.from_id("landsat:LC08:PRE:TOAR:meta_LC80270312016188_v1")
        >>> arr = scene.ndarray("red green blue", ctx)
        >>> type(arr)
        <class 'numpy.ma.core.MaskedArray'>
        >>> arr.shape
        (3, 15259, 15340)
        >>> red_band = arr[0]

        Raises
        ------
        ValueError
            If requested bands are unavailable.
            If band names are not given or are invalid.
            If the requested bands have different dtypes.
        NotFoundError
            If a Scene's id cannot be found in the Descartes Labs catalog
        BadRequestError
            If the Descartes Labs platform is given invalid parameters
        """
        if raster_client is None:
            raster_client = Raster()

        if not (-3 < bands_axis < 3):
            raise ValueError("Invalid bands_axis; axis {} would not exist in a 3D array".format(bands_axis))

        bands = self._bands_to_list(bands)
        common_data_type = self._common_data_type_of_bands(bands)

        self_bands = self.properties["bands"]
        if mask_alpha:
            if "alpha" not in self_bands:
                raise ValueError(
                    "Cannot mask alpha: no alpha band for the product '{}'".format(self.properties["product"])
                )
            try:
                alpha_i = bands.index("alpha")
            except ValueError:
                bands.append("alpha")
                drop_alpha = True
            else:
                if alpha_i != len(bands) - 1:
                    raise ValueError("Alpha must be the last band in order to reduce rasterization errors")
                drop_alpha = False

        raster_params = ctx.raster_params
        full_raster_args = dict(
            inputs=self.properties["id"],
            order="gdal",
            bands=bands,
            scales=None,
            data_type=common_data_type,
            **raster_params
        )

        try:
            arr, info = raster_client.ndarray(**full_raster_args)
        except NotFoundError as e:
            raise NotFoundError("'{}' does not exist in the Descartes catalog".format(self.properties["id"]))
        except BadRequestError as e:
            msg = ("Error with request:\n"
                   "{err}\n"
                   "For reference, dl.Raster.ndarray was called with these arguments:\n"
                   "{args}")
            msg = msg.format(err=e, args=json.dumps(full_raster_args, indent=2))
            six.raise_from(BadRequestError(msg), None)

        if len(arr.shape) == 2:
            # if only 1 band requested, still return a 3d array
            arr = arr[np.newaxis]

        if mask_nodata or mask_alpha:
            if mask_alpha:
                alpha = arr[-1]
                if drop_alpha:
                    arr = arr[:-1]
                    bands.pop(-1)

            mask = np.zeros_like(arr, dtype=bool)

            if mask_nodata:
                for i, bandname in enumerate(bands):
                    nodata = self_bands[bandname].get('nodata')
                    if nodata is not None:
                        mask[i] = arr[i] == nodata

            if mask_alpha:
                mask |= alpha == 0

            arr = np.ma.MaskedArray(arr, mask, copy=False)

        if bands_axis != 0:
            arr = np.moveaxis(arr, 0, bands_axis)
        if raster_info:
            return arr, info
        else:
            return arr

    @property
    def __geo_interface__(self):
        # QUESTION: this returns a Geometry, should it be a Feature and include properties?
        try:
            return self.geometry.__geo_interface__
        except AttributeError:
            return self.geometry

    def _dict(self):
        return dict(
            geometry=self.__geo_interface__,
            properties=self.properties,
        )

    def _repr_json_(self):
        return self._dict()

    def __repr__(self):
        return repr(self._dict())

    def _common_data_type_of_bands(self, bands):
        "Ensure all requested bands are available, and that they all have the same dtypes"
        self_bands = self.properties["bands"]
        common_data_type = None
        for b in bands:
            try:
                band = self_bands[b]
            except KeyError:
                msg = "Band '{}' is not available in the product '{}'".format(b, self.properties["product"])
                # check if they maybe wanted a derived band but forgot the ``derived:`` prefix
                as_derived = "derived:" + b
                if as_derived in self_bands:
                    msg += ", but '{}' is. Did you mean that?".format(as_derived)
                six.raise_from(ValueError(msg), None)
            try:
                data_type = band["dtype"]
            except KeyError:
                six.raise_from(
                    ValueError(
                        "Band '{}' of product '{}' has no 'dtype' field. "
                        "If you created this product, you can fix the metadata "
                        "at https://catalog.descarteslabs.com.".format(b, self.properties["product"])
                    ),
                    None
                )
            if common_data_type is None:
                common_data_type = data_type
            else:
                if data_type != common_data_type:
                    raise ValueError(
                        "Bands must all have the same dtype. The band '{}' has dtype '{}', "
                        "but all bands before it had the dtype '{}'.".format(b, data_type, common_data_type)
                    )
        return common_data_type

    @staticmethod
    def _bands_to_list(bands):
        if isinstance(bands, six.string_types):
            return bands.split(" ")
        if not isinstance(bands, (list, tuple)):
            raise TypeError("Expected list or tuple of band names, instead got {}".format(type(bands)))
        if len(bands) == 0:
            raise ValueError("No bands specified to load")
        return list(bands)

    @staticmethod
    def _scenes_bands_dict(metadata_bands):
        """
        Convert bands dict from metadata client ({id: band_meta})
        to {<name, or ID if derived>: band_meta}
        """
        return DotDict({
            id if id.startswith("derived") else meta["name"]: meta
            for id, meta in six.iteritems(metadata_bands)
        })
